/* header for anything that wants to use the 16C84's serial programming mode */
#ifndef PROG84_H
#define PROG84_H
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifndef __MSDOS__
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#endif
#include "io_ports.h"


#ifndef	EXTERN
#define	EXTERN extern
#endif

/* MSDOS needs SLEEP100ns() */
#if defined( __MSDOS__ ) 
# define SLEEP100ns(usec) sleep100ns(usec)
#else
# define SLEEP100ns(usec) delay_ns(100)
#endif

/***
 *** Externally visible variables & functions
 ***/

/*** VARIABLES ***/

    /***
     *** Bit indicators for the various control & feedback signals.
     *** See lp_io for what a bit indicator is.
     ***/
extern int p_power, p_mclr, p_vpp2, p_vpp3, p_data_d,
           p_clock, p_data, p_clock_f, p_data_f;
extern int p_loopcnt, p_progtype, p_no_power_C4C8;

    /*** lp_io control structure ***/
extern struct lp_io lpb;

    /*** name of file to read pin assignments from; defaults to lp_cfg ***/
extern char *cfg_file;

    /*** path to the called program - if possible */
extern char *progpath;

    /*** set this to nonzero for copious output ***/
extern int p84_verbose;
EXTERN int verbose;
EXTERN int emulate_write;


/*** FUNCTIONS ***/

void set_progpath(char *prog) ;

    /***
     *** set up and shut down the programmer, as well as the software layers
     *** under this (lpb_cfg_read, open_io(), etc.)
     ***/
void delay_ns(unsigned long dly);
void progSetup();
void progShutdown(); /* powers down */
void power_and_reset();

    /*** simple hardware test sequence. called by progSetup().
     *** returns non-zero on problem.
     ***/
int testHW();

/*** Routines to clock some bits through the chip serially. */
    /*** send bits through the '84 */
unsigned int pipeBits(unsigned int, int);

    /*** send bits, confirm, exit on error ***/
void xPipeBits(unsigned int, int);

    /*** send command & 14 data bits ***/
void genericWrite(int, unsigned int);

    /*** send command, then read 14 data bits ***/
unsigned int genericRead(int);

    /*** print a list of supported devices ***/
void printListSupported();

/* ************************************************************************ */
/* Defines for the descriptive programmer.                                  */

/* MICROCHIP'S SERIAL COMMANDS */
#define SC_LOADCFG	 0	/* DATA */
#define SC_CMD1		 1	/* used to clear protection */
#define SC_LOADPROG	 2	/* DATA */
#define SC_LOADDATA	 3	/* DATA */
#define SC_READPROG	 4	/* DATA */
#define SC_READDATA	 5	/* DATA */
#define SC_INCREMENT	 6
#define SC_CMD7		 7	/* used to clear protection */
#define SC_PROG	 	 8
#define SC_ERASEPGM	 9
#define SC_ENDPROG2	10	/* end programming externally timed */
#define SC_ERASEDATA	11
#define SC_CMD11	11	/* ??? */
#define SC_ENDPROG	14	/* needed for 16c6x/7x/9xx programming */
#define SC_ERASEROW	17	/* erase row internally timed */
#define SC_PROG2	18	/* begin programming externally timed */
#define SC_ENDPROG3	23	/* 16F877A, etc. */
#define SC_PROG3   	24	/* Programm only without erase 16F877A */
#define SC_ERASECHIP	31

/* Bit-twiddling macros, for convenience */
#define SET(bitname, value) lpb_write(&lpb, p_ ## bitname, value)
/* Before using GET on new data from port, use REFRESH to read port !! 
   FLUSH may not always have appropiate delay before reading port */
#define GET(bitname) lpb_test(&lpb, p_ ## bitname)
#define FLUSH do{ lpb_flush(&lpb); SLEEP100ns(1); lpb_refresh(&lpb); } while(0)
/* old: #define FLUSH do{ lpb_flush(&lpb); lpb_refresh(&lpb); } while(0) */
#define REFRESH do { lpb_refresh(&lpb); } while(0)

/* Commands that write */
#define loadConfiguration(cfg) genericWrite(SC_LOADCFG, cfg)
#define loadProg(data)         genericWrite(SC_LOADPROG, data)
#define loadData(data)         genericWrite(SC_LOADDATA, data)

/* Commands that read */
#define readProg()	genericRead(SC_READPROG)
#define readData()	genericRead(SC_READDATA)

/* Commands that neither read nor write */
#define incAddr()	xPipeBits(SC_INCREMENT, 6)

/* EEprom/i2c defines */
/* actually it should be named EE_ACK/EE_NOACK .*/
#define I2C_ACK 0
#define I2C_NOACK 1
#define EE_CTRL_WR 0xa0
#define EE_CTRL_RD 0xa1

/* EEprom/i2c Bit-twiddling macros, for convenience */
#define SET(bitname, value) lpb_write(&lpb, p_ ## bitname, value)
#define GET(bitname) lpb_test(&lpb, p_ ## bitname)

#define I2C_SET(ctrl,dat)  i2c_out(ctrl,dat)
#define I2C_GET()   lpb_test(&lpb, p_data_f)
#define I2C_DEBUG(x)  if (i2c_debug) (x)

typedef char BOOL;
#ifdef __MSDOS__
typedef unsigned WORD;
typedef long WORD32;
#else
typedef unsigned short WORD;
typedef int WORD32;
#endif

enum {  
  picid_addr = 0x2000, /* start of picid address */
  fuse_addr  = 0x2007, /* start of fuse address */
  data_addr  = 0x2100  /* start og eeprom data in PIC */
};

enum tPromType { EEPROM, EPROM, FLASH1, FLASH2, FLASH3, FLASH4, FLASH5 };

typedef struct
{
  char	       *name;		// chip name e.g. 16F84
  enum tPromType type;		// 0=Eprom, 1=early flash, 2=16f84A & later flash
  WORD		core_bits;	// core size ie 12 or 14 bits
  WORD		page_size;	// Size of memory page for later Flash types
  WORD		prg_dly;	// time for 1 program pulse in us.
				// For flash it is MAX time. For all flash
				// devices, bulk erase time is 10ms set in
				// erase routine
  WORD		n_tries;	// (for eprom) Number of initial
				// program pulses
  WORD		n_reburn;	// (for eprom) The overprogram ratio
  WORD		prog_size;	// Program memory size in words
  WORD		eeprom_size;	// Data memory size in words
  WORD		dev_id;		// the device id (ie hi) bits,
  WORD		rev_bits;	// mask for device id
				// (ie how many revision bits)
} chip_info;

extern chip_info *chip_p;

chip_info * get_chip_p(char *chname) ;
int set_chip_p(char *chname);

/* I2C for EERPOM */

extern int i2c_debug;

void i2c_start() ;
void i2c_stop() ;
int i2c_sendbyte(unsigned char dat);
unsigned char i2c_readbyte(int ack);
void i2c_seqread(WORD adr, unsigned char *data, int len) ;
unsigned char i2c_read(WORD adr) ;
int i2c_waitack(unsigned int t);
void i2c_pagewrite(WORD adr, unsigned char *data, WORD PageSize) ;
void i2c_write(WORD adr, WORD data) ;


#endif /* PROG84_H */
