; (C) 1999 Frank Damgaard - GNU -CopyLeft
;
;                TITLE           "thru02.asm"
;                LIST           P=16C84
; This is compiled with picasm  http://www.iki.fi/trossi/pic/
; picasm should compile with most ansi-c compilers

                opt list
                device pic16c84
                config CP=off,OSC=xt

                IO_PIN  EQU     007H    ;SET INPUT / OUTPUT PIN
                IO_MASK EQU     07FH

		INDIR	EQU	00H	; Indirect pointer
		RTCC	EQU	01H	; Real time clock / counter
		PCL	EQU	02H	; Program counter low byte
		STATUS	EQU	03H	; Status register
		CY	EQU	0	; Carry/borrow bit
		DC	EQU	1	; Digit carry
		Z	EQU	2	; Zero flag
		PD	EQU	3	; Power down
		T0	EQU	4	; Time-out bit
		RP0	EQU	5	; Register page select bits
		RP1	EQU	6	; Not used in 16C84
		RP2	EQU	7	; Not used in 16C84
		FSR	EQU	04H	; Indirect pointer
		PORTA	EQU	05H	; PORT A
		PORTB	EQU	06H	; PORT B
		EEDATA	EQU	08H	; Eeprom data
		EEADR	EQU	09H	; Eeprom address
		PCLATH	EQU	0AH	; Program counter high byte
		INTCON	EQU	0BH	; Interrupt control
		GIE	EQU	07H	;  global interrupt enable bit
;  if picasm doesn't like 08?H for register addresses (TRISA/B EECON1/2)
;  change to 05H, 06H, 08H and 09H :
		OPTION_REG EQU  081H
		TRISA	EQU	085H	; Data direction port A
		TRISB	EQU	086H	; Data direction port B
		EECON1	EQU	088H	; Eeprom control
		EECON2	EQU	089H	; Eeprom control
		RD	EQU	0	;  eeprom rd bit
		WR	EQU	1	;  eeprom wr bit
		WREN	EQU	2	;  eeprom wr enable bit
		WRERR	EQU	3	;  eeprom ....
		EEIF	equ	4	;  eeprom ....
		W	EQU	0	; W reg. as destination
		F	EQU	1	; F file as destination

		TMP1	equ	0CH	; reg 0ch tmp
		BITCNT  equ	0DH	; bit#
		ADDRL	equ	0EH	; eeprom address
		ADDRH	equ	0FH	; eeprom address
		RCVSTAT	equ	10H	; RCVSTAT bit 0:= last byte to tx 
					;         bit 1:= last byte to rx

                SCLK    EQU     5       ;SERIAL EEPROM CLOCK LINE
                SDATA   EQU     4       ;SERIAL EEPROM DATA LINE


		ORG	000H
;
; Set data=rb7 = high (1), clk=0 
; after reset send clockpulses until RB7=data goes low ( = ack)

; Sending data to the PIC:
; send the specified number of clockpulses 
;    the first probe of RB7 occurs after 40 clockpulses
;    each new bit is probed after 28 clocks (8 bit data + 1 continue/ack bit)
;    send some  clock pulses and wait for "ack" = RB7=data goes low.
;    to ensure proper setup times: 
;    1) wait for "ack"
;    2) send 4-8 clocks (only for word 0 and word 1)
;    3) repeat for all 8 bits:
;           set RB7 to value of the i'th data bit
;           send 28 clocks
;    4) if this is the last byte sent in this block the 9'th bit is 1
;       else it is 0. Set RB7 to this value and send 28 clocks
;    5) send 40 clockpulses
;    6) goto 1)
;
; The bytes sent have following format : 
; word 0:  controlbyte, selects read/write and includes the high 3 bits of
;          the address. :  1010aaaR0 , R=1 if read data, R=0 if write data
; word 1:  lower 8bit of the address (24c16 eeprom): aaaaaaaa0
; word 2:  first  byte of data dddddddd0 (at address A)
; word 3:  second byte of data dddddddd0 (at address A+1)
; .....
; word 2+N:  last byte of data dddddddd1 (at adress A+N)
;                                      ^ signals last byte
; Reading data from the PIC:
;    Almost like sending data, send word 0 = command= 1010aaaR0=1010aaa10
;    Send lower 8 bit of address in word 1
;    Received bytes are only 8 bit wide.
;    Bytes will be sent by thru02 as long as clocks applied
;    and until MCLR is applied.
;
;    1) wait for "ack"
;    2) send 4-8 clocks
;    3) repeat for all 8 bits:
;           set RB7 to value of the i'th data bit
;           send 28 clocks
;    4) if this is the last byte sent in this block the 9'th bit is 1
;       else it is 0. Set RB7 to this value and send 28 clocks
;    5) send 40 clockpulses
;    6) do 1)-5) for word 0 and word 1
;    7) send 96 clocks
;    8) repeat for all 8 bits (i=0..7):
;           read bit value on RB7 of the i'th data bit
;           send 28 clocks
;    9) send 16 clocks
;   10) wait for "ack".

;-----------------------------------------------------------------------------
;  Program Start	Send Atr and Read data from decoder
;-----------------------------------------------------------------------------
START
		clrf	PORTA
		bsf	STATUS,RP0
		movlw	0x1f
		movwf	TRISA
INTXX           bcf	INTCON,GIE	; disable interrupt 
		movlw	0xff
		movwf	TRISB
		movlw	0x7f
		movwf	OPTION_REG
		bcf	OPTION_REG,7
		bcf     STATUS,RP0

SLOOP		; read until controlbyte 0xa*  is received
		; high 3 bits of address in bit 1-3
		call    txack    ; (clocks) +48
		call    rxbyte   ; (clocks) +0,28,56,84,112,140,168,196
		movwf   ADDRH
		andlw   0xf0
		xorlw   0xa0
		btfss	STATUS,Z
		goto	SLOOP	; error wrong coltrolbyte try again
		btfsc	ADDRH,0
		goto	RDMODE
 if 1
WRMODE		; faster page write , approx. 2x times faster
		call	RDADDR
		clrf 	RCVSTAT		
WRNEXT		;                ;do {  /* only max 16 byte page write */
		call	WRITE_EE ; write to eeprom 
WRNEXTB		;                ;  do {
		call	txack	 ;   (clocks) +48
		call    rxbyte	 ;   (clocks) +0,28,56,84,112,140,168,196 +28
		movwf	EEDATA
		call	WRITE_BYTE
		btfsc	RCVSTAT,0     ; if last byte goto SLOOP
		goto    SLOOP
		incf	ADDRL,F	      ; W= (addrh,addrl)++
		btfsc	STATUS,Z
		incf	ADDRH
		movf	ADDRL,W	      
		andlw	0fH      
		btfsc	STATUS,Z ;  while ( addrl&0xf >0 )
		goto	PAGELAST
		goto    WRNEXTB     
PAGELAST	; last byte in 16 byte page boundary (0x0...0xf)
		call BSTOP	 ;  send BSTOP
		movf	ADDRL,W
		goto WRNEXT      ;while ( 1 )
 else
WRMODE		; older slower write
		call	RDADDR
		clrf 	RCVSTAT		
WRNEXT
		call	WRITE_EE ; write to eeprom 
		call	txack	 ; (clocks) +48
		call    rxbyte	 ; (clocks) +0,28,56,84,112,140,168,196 +28
		movwf	EEDATA
		call	WRITE_BYTE
		btfsc	RCVSTAT,0
		goto    SLOOP
		incf	ADDRL,F		; W= (addrh,addrl)++
		btfsc	STATUS,Z
		incf	ADDRH
		movf	ADDRL,W		; 
		goto    WRNEXT
 endif

 if 1
RDMODE		; faster sequentiel read from eeprom
		; send controlbyte + address
		; for each read byte, only read one byte from eeprom
		; should be approx. 3 times faster
		call	RDADDR	 ; read low part address and adjust ADDRH
;		clrf	RCVSTAT  ; not last byte
		bsf	RCVSTAT,0 ; ony byte at a time
		call	READ_EE	 ; read from eeprom addr=ADDRH/L
		goto    RDNEXTC
RDNEXT                           ; do {
	;	call	RX	 ;   read next byte from eeprom
		call    READ_EE  ;   (value in EEDATA and W

RDNEXTC		
		call	txack	 ;   (clocks) +48
		movf	EEDATA,W ;   <= data read should now be in W
;		movlw   0xe5	 ;   1. test (always send 0xe5)
;		movf	ADDRL,W  ;   2. test (sendback current address (low)
		call    txbyte	 ;   (clocks) 
		incf	ADDRL,F	 ;    W= (addrh,addrl)++
		btfsc	STATUS,Z
		incf	ADDRH
		movf	ADDRL,W	 ; 
		goto    RDNEXT	 ; } forever;

ENDLP1		
		call	txack
		nop
		nop
		goto ENDLP1

 else
RDMODE		; older slower read, 
		; for each read byte: send address to eeprom and read byte
		call	RDADDR	 ; read low part address and adjust ADDRH
RDNEXT                          ; do {
		call	READ_EE	 ;   read from eeprom
		movwf	EEDATA
		call	txack	 ;   (clocks) +48
		movf	EEDATA,W
		call    txbyte	 ;   (clocks) 
		incf	ADDRL,F	 ;    W= (addrh,addrl)++
		btfsc	STATUS,Z
		incf	ADDRH
		movf	ADDRL,W	 ; 
		goto    RDNEXT	 ; } forever;
		goto 	ENDLP1
ENDLP1		
		call	txack
		nop
		nop
		goto ENDLP1

 endif

; Read address
; ADDRH = xxxAAAx, upper 3 bits of eeprom address = AAA
; returns low part of address in ADDRL=EEADR=W
;         hight part in ADDRH (bit 2-0)
RDADDR		rrf	ADDRH,F
		movlw	0x07
		andwf	ADDRH,F
		call	txack	 ; (clocks) +48
		; read low part (8 bits) of address
		call    rxbyte	 ; (clocks) +0,28,56,84,112,140,168,196
		movwf	EEADR		; since writing 16byte blocks
		movwf	ADDRL
		return

; Transmit "ack" (set RB7 output to "low" for at most 4 clocks)
txack		; uses 4*7+4 (incl call +8)= 32/40
		clrf	PORTB
		bsf	STATUS,RP0	; bank 1
		bcf	TRISB,IO_PIN	; Define PORT B I/O b7=out
		bcf	STATUS,RP0	; bank 0
 		bcf	PORTB,IO_PIN	; set RB7=0
		bsf	PORTB,IO_PIN	; set RB7=1
		bsf	STATUS,RP0	; bank 1
		bsf	TRISB,IO_PIN	; back in rx-mode
		bcf	STATUS,RP0	; bank 0
		return

; returns reg W = read byte
; bits read in acsending order : 0,1,2,3,4,5,6,7 + ACK/CONT-bit
rxbyte		clrf	TMP1	; holds result
		movlw	01h	
		movwf	BITCNT	; holds bitmask (01, 02, 04, ...80)
rx_loop				; do {
		movf	BITCNT,W   ;   W=bitcnt;
		BTFSC	PORTB,IO_PIN ; if ( set(PORTB,IO_PIN)
		iorwf	TMP1,F     ;     tmp1 |= W;
		addwf	BITCNT,F   ;   bitcnt += W /* bitcnt <<=1 */
		btfss	STATUS,CY ; } while ( bitcnt <=80h)
		goto	rx_loop
		movf	TMP1,W	
		BTFSC	PORTB,IO_PIN	; test continue/ack-bit
		bsf	RCVSTAT,0
		return

; write W to RB7 with bits in acsending order
txbyte		movwf	EEDATA
		clrf	PORTB
		bsf	STATUS,RP0	; bank 1
		bcf	TRISB,IO_PIN	; Define PORT B I/O RB7=output
		bcf	STATUS,RP0	; bank 0
		bcf	STATUS,CY
		rrf	EEDATA,F	; rotate(eedata,cy)
		movlw	8		; for ( t=8; t>0 ; 8--) {
		movwf	TMP1
txbloop
		rrf	EEDATA,F	;  rotate(eedata,cy)
		movlw	IO_MASK		;  outb(portb, eedata&IO_MASK)
		iorwf	EEDATA,W
		movwf	PORTB	
		decfsz	TMP1,F
		goto	txbloop		; }
		bsf	PORTB,IO_PIN	;  outp(portb,0xff)
		bsf	STATUS,RP0	; bank 1
		bsf	TRISB,IO_PIN	; RB7 back in input-mode
		bcf	STATUS,RP0	; bank 0
		return

;
; following code borrowed from multi215.asm +
; - read modified to do sequential read 
;
;######################################################
;	EEPROM READ ROUTINE
;######################################################
;		Used reg: EEADR, EEDATA,RCVSTAT,W 
;		IN  : W= READ ADRESS
;		OUT : W= BYTE

READ_EE 	MOVWF	EEADR	       ; store byte in reg EEADR

READ_EE_DIR	CALL	BSTART		; generate start bit
		BCF	STATUS,CY	; WRITE SEQUENCE
		CALL	CONTROL_BYTE	; CALC CONTROL BYTE AND TRANSMIT
		MOVF	EEADR,W 	; get word address....
READ_EE1	CALL	TX		; and send it

READ_CURRENT	CALL	BSTART		; generate start bit
		BSF	STATUS,CY	; READ SEQUENCE
		CALL	CONTROL_BYTE	; CALC CONTROL BYTE AND TRANSMIT
		CALL	RX		; read 1 byte from serial EE
		btfsc	RCVSTAT,1	; if lastbyte then
		CALL	BSTOP		;   send stop bit to end transmission
		MOVF	EEDATA,W	; GET RECEIVED BYTE
		RETURN

;######################################################
;	EEPROM WRITE ROUTINE
;######################################################
;		USED REG : EEADR, EEDATA, W, AADRH
;		IN	 : EEADR = write adress.

WRITE_EE	MOVWF	EEADR
WRITE_EE_DIR	CALL	BSTART		;generate start bit
		BCF	STATUS,CY	;WRITE SEQUENCE
		CALL	CONTROL_BYTE	;CALC CONTROL BYTE AND TRANSMIT
		MOVF	EEADR,W 	;GET WORD ADRESS
		GOTO	TX		;WRITE @ ADRESS COUNTER

WRITE_BYTE	CALL	TX		;WRITE BYTE (W) @ ADRESS COUNTER
		btfsc	RCVSTAT,0	;if lastbyte then
		GOTO	BSTOP		;  END READ SEQUENCE
		RETURN

;**********************************************
;	GENERATE CONTROL BYTE
;**********************************************
CONTROL_BYTE	RLF	ADDRH,W		; => CY=1
		ANDLW	B'00001111'     ;MASK OUT UPPER NIBBLE
		IORLW	B'10100000'     ;COMPLETE BYTE 1010AAAD
					;A=ADRESS D=DIRECTION

					;SEND CONTROL BYTE


;**********************************************
;	TRANSMIT DATA SUBROUTINE
;**********************************************

TX		MOVWF	EEDATA		;STORE BYTE TO TX
		MOVLW	B'11001111'     ; set data,clock as outputs
		TRIS	PORTB

		CALL    TX4		;TRANSMIT 8 TIMES
		CALL    TX4
		GOTO 	BITIN

TX4		CALL    TXLP
		CALL    TXLP
		CALL    TXLP
TXLP	 	RLF	EEDATA,F	; ROTATE BIT TO CARRY
                GOTO    BITOUT_R        ; send the bit to serial EE

;**********************************************
;	RECEIVE DATA SUBROUTINE
;**********************************************
RX   		MOVLW	B'11011111'    ; make SDATA an input line/SCLK = OUT
		TRIS	PORTB
		CALL    RX4
		CALL    RX4
		btfss   RCVSTAT,1      ; if not last byte then		
		bcf	STATUS,CY      ;   set Ack=0 
		btfsc	RCVSTAT,1      ; else	
	        BSF     STATUS,CY      ;   SET ACK BIT=1
                GOTO    BITOUT         ; TO FINISH TRANSMISSION

RX4		CALL	RXLP
     		CALL 	RXLP
		CALL	RXLP
RXLP            CALL    BITIN_R         ; READ A BIT
	 	RLF	EEDATA,F	; SHIFT CARRY TO BYTE
   		RETURN


;*********************************************
;	START BIT SUBROUTINE
;*********************************************
BSTART		BSF	PORTB,SDATA	; make sure data is high
		MOVLW	B'11001111'
		TRIS	PORTB		; set data and clock lines for output
		BSF	PORTB,SCLK	; set clock high
		NOP
		BCF	PORTB,SDATA	; data line goes low during
		GOTO	BC_END		; high clock for start bit

;**********************************************
;	STOP BIT SUBROUTINE
;**********************************************

BSTOP		BCF	PORTB,SDATA	; make sure data line is low
		MOVLW	B'11001111'
		TRIS	PORTB		; set data/clock lines as outputs
		NOP
		BSF	PORTB,SCLK	; set clock high
		NOP
		BSF	PORTB,SDATA	; data goes high while clock high
					; for stop bit
B_END		NOP
BC_END		BCF	PORTB,SCLK	; set clock low again
		RETURN

;**********************************************
;	BITOUT ROUTINE , SEND BIT WHEN CARRY SET
;**********************************************

BITOUT		MOVLW	B'11001111'     ; set data,clock as outputs
		TRIS	PORTB
BITOUT_R	BTFSS	STATUS,CY	; check for state of data bit to xmit
		BCF	PORTB,SDATA	; set data line low
		BTFSC	STATUS,CY
		BSF	PORTB,SDATA	; high? set data line high

CLKOUT		BSF	PORTB,SCLK	; set clock line high
		GOTO	B_END

;**********************************************
;	BITIN ROUTINE
;**********************************************
BITIN		MOVLW	 B'11011111'    ; make SDATA an input line/SCLK = OUT
		TRIS	 PORTB
		bcf	OPTION_REG,7	; make sure weak pull-up is on
BITIN_R		BSF	 PORTB,SCLK	; set clock line high
		NOP
		BSF	 STATUS,CY	; assume input bit is high
		BTFSS	 PORTB,SDATA	; read the data bit
		BCF	 STATUS,CY	; input bit was low ,clear CARRY
		GOTO	 BC_END


		END
